#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#

import datetime
import os
import os.path
import pickle
import shutil
import tempfile
import unittest

import httpretty
import requests
import rq
from dateutil.tz import UTC

from arthur import __version__
from arthur.errors import NotFoundError
from arthur.jobs import (JobResult,
                         PercevalJob,
                         execute_perceval_job)
from grimoirelab_toolkit.datetime import datetime_utcnow
from perceval.archive import ArchiveManager

from base import TestBaseRQ


BUGZILLA_SERVER_URL = 'http://example.com'
BUGZILLA_BUGLIST_URL = BUGZILLA_SERVER_URL + '/buglist.cgi'
BUGZILLA_BUG_URL = BUGZILLA_SERVER_URL + '/show_bug.cgi'
BUGZILLA_BUG_ACTIVITY_URL = BUGZILLA_SERVER_URL + '/show_activity.cgi'

REDMINE_URL = 'http://example.com'
REDMINE_ISSUES_URL = REDMINE_URL + '/issues.json'
REDMINE_ISSUE_2_URL = REDMINE_URL + '/issues/2.json'
REDMINE_ISSUE_5_URL = REDMINE_URL + '/issues/5.json'
REDMINE_ISSUE_9_URL = REDMINE_URL + '/issues/9.json'
REDMINE_ISSUE_7311_URL = REDMINE_URL + '/issues/7311.json'
REDMINE_USER_3_URL = REDMINE_URL + '/users/3.json'
REDMINE_USER_4_URL = REDMINE_URL + '/users/4.json'
REDMINE_USER_24_URL = REDMINE_URL + '/users/24.json'
REDMINE_USER_25_URL = REDMINE_URL + '/users/25.json'

REDMINE_URL_LIST = [
    REDMINE_ISSUES_URL, REDMINE_ISSUE_2_URL, REDMINE_ISSUE_5_URL,
    REDMINE_ISSUE_9_URL, REDMINE_ISSUE_7311_URL, REDMINE_USER_3_URL,
    REDMINE_USER_4_URL, REDMINE_USER_24_URL, REDMINE_USER_25_URL
]


def read_file(filename, mode='r'):
    dir = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(dir, filename), mode) as f:
        content = f.read()
    return content


def setup_mock_bugzilla_server():
    """Setup a mock Bugzilla server for testing"""

    http_requests = []
    bodies_csv = [read_file('data/bugzilla_buglist.csv'),
                  read_file('data/bugzilla_buglist_next.csv'),
                  ""]
    bodies_xml = [read_file('data/bugzilla_version.xml', mode='rb'),
                  read_file('data/bugzilla_bugs_details.xml', mode='rb'),
                  read_file('data/bugzilla_bugs_details_next.xml', mode='rb')]
    bodies_html = [read_file('data/bugzilla_bug_activity.html', mode='rb'),
                   read_file('data/bugzilla_bug_activity_empty.html', mode='rb')]

    def request_callback(method, uri, headers):
        if uri.startswith(BUGZILLA_BUGLIST_URL):
            body = bodies_csv.pop(0)
        elif uri.startswith(BUGZILLA_BUG_URL):
            body = bodies_xml.pop(0)
        else:
            body = bodies_html[len(http_requests) % 2]

        http_requests.append(httpretty.last_request())

        return 200, headers, body

    httpretty.register_uri(httpretty.GET,
                           BUGZILLA_BUGLIST_URL,
                           responses=[
                               httpretty.Response(body=request_callback)
                               for _ in range(3)
                           ])
    httpretty.register_uri(httpretty.GET,
                           BUGZILLA_BUG_URL,
                           responses=[
                               httpretty.Response(body=request_callback)
                               for _ in range(3)
                           ])
    httpretty.register_uri(httpretty.GET,
                           BUGZILLA_BUG_ACTIVITY_URL,
                           responses=[
                               httpretty.Response(body=request_callback)
                               for _ in range(7)
                           ])

    return http_requests


def setup_mock_redmine_server(max_failures=0):
    """Setup a mock Redmine HTTP server"""

    http_requests = []
    failures = max_failures

    issues_body = read_file('data/redmine/redmine_issues.json', 'rb')
    issues_next_body = read_file('data/redmine/redmine_issues_next.json', 'rb')
    issues_empty_body = read_file('data/redmine/redmine_issues_empty.json', 'rb')
    issue_2_body = read_file('data/redmine/redmine_issue_2.json', 'rb')
    issue_5_body = read_file('data/redmine/redmine_issue_5.json', 'rb')
    issue_9_body = read_file('data/redmine/redmine_issue_9.json', 'rb')
    issue_7311_body = read_file('data/redmine/redmine_issue_7311.json', 'rb')
    user_3_body = read_file('data/redmine/redmine_user_3.json', 'rb')
    user_4_body = read_file('data/redmine/redmine_user_4.json', 'rb')
    user_24_body = read_file('data/redmine/redmine_user_24.json', 'rb')
    user_25_body = read_file('data/redmine/redmine_user_25.json', 'rb')

    def request_callback(method, uri, headers):
        nonlocal failures

        status = 200
        last_request = httpretty.last_request()
        params = last_request.querystring

        if uri.startswith(REDMINE_ISSUES_URL):
            updated_on = params['updated_on'][0]
            offset = params['offset'][0]

            if updated_on == '>=1970-01-01T00:00:00Z' and offset == '0':
                body = issues_body
            elif updated_on == '>=1970-01-01T00:00:00Z' and offset == '3':
                body = issues_next_body
            elif updated_on == '>=2016-07-27T00:00:00Z' and offset == '0':
                body = issues_next_body
            elif updated_on == '>=2011-12-08T17:58:37Z' and offset == '0':
                body = issues_next_body
            else:
                body = issues_empty_body
        elif uri.startswith(REDMINE_ISSUE_2_URL):
            body = issue_2_body
        elif uri.startswith(REDMINE_ISSUE_5_URL):
            body = issue_5_body
        elif uri.startswith(REDMINE_ISSUE_9_URL):
            body = issue_9_body
        elif uri.startswith(REDMINE_ISSUE_7311_URL):
            if failures > 0:
                status = 500
                body = "Internal Server Error"
                failures -= 1
            else:
                body = issue_7311_body
        elif uri.startswith(REDMINE_USER_3_URL):
            body = user_3_body
        elif uri.startswith(REDMINE_USER_4_URL):
            body = user_4_body
        elif uri.startswith(REDMINE_USER_24_URL):
            body = user_24_body
        elif uri.startswith(REDMINE_USER_25_URL):
            body = user_25_body
        else:
            raise

        http_requests.append(last_request)

        return status, headers, body

    for url in REDMINE_URL_LIST:
        httpretty.register_uri(httpretty.GET,
                               url,
                               responses=[
                                   httpretty.Response(body=request_callback)
                               ])

    return http_requests


class TestJobResult(unittest.TestCase):
    """Unit tests for JobResult class"""

    def test_job_result_init(self):
        result = JobResult('1234567890', 8, 'mytask',
                           'mock_backend', 'category')

        self.assertEqual(result.job_id, '1234567890')
        self.assertEqual(result.job_number, 8)
        self.assertEqual(result.task_id, 'mytask')
        self.assertEqual(result.backend, 'mock_backend')
        self.assertEqual(result.category, 'category')
        self.assertEqual(result.summary, None)

    def test_to_dict(self):
        """Test whether a JobResult object is converted to a dict"""

        result = JobResult('1234567890', 8, 'mytask',
                           'mock_backend', 'category')

        expected = {
            'job_id': '1234567890',
            'job_number': 8,
            'task_id': 'mytask'
        }

        d = result.to_dict()
        self.assertEqual(d, expected)


class TestPercevalJob(TestBaseRQ):
    """Unit tests for PercevalJob class"""

    def setUp(self):
        self.tmp_path = tempfile.mkdtemp(prefix='arthur_')
        super().setUp()

    def tearDown(self):
        shutil.rmtree(self.tmp_path)
        super().tearDown()

    def test_init(self):
        """Test the initialization of the object"""

        job = PercevalJob('1234567890', 8, 'mytask', 'git', 'commit',
                          self.conn, 'items')

        self.assertEqual(job.job_id, '1234567890')
        self.assertEqual(job.job_number, 8)
        self.assertEqual(job.task_id, 'mytask')
        self.assertEqual(job.backend, 'git')
        self.assertEqual(job.category, 'commit')
        self.assertEqual(job.conn, self.conn)
        self.assertEqual(job.qitems, 'items')
        self.assertEqual(job.archive_manager, None)

        result = job.result
        self.assertIsInstance(job.result, JobResult)
        self.assertEqual(result.job_id, '1234567890')
        self.assertEqual(job.job_number, 8)
        self.assertEqual(result.task_id, 'mytask')
        self.assertEqual(result.backend, 'git')
        self.assertEqual(job.category, 'commit')
        self.assertEqual(result.summary, None)

    def test_backend_not_found(self):
        """Test if it raises an exception when a backend is not found"""

        with self.assertRaises(NotFoundError) as e:
            _ = PercevalJob('1234567890', 8, 'mytask',
                            'mock_backend', 'acme-category',
                            self.conn, 'items')
            self.assertEqual(e.exception.element, 'mock_backend')

    def test_run(self):
        """Test run method using the Git backend"""

        job = PercevalJob('1234567890', 8, 'mytask',
                          'git', 'commit',
                          self.conn, 'items')
        args = {
            'uri': 'http://example.com/',
            'gitpath': os.path.join(self.dir, 'data/git_log.txt')
        }
        archive_args = {
            'archive_path': self.tmp_path,
            'fetch_from_archive': False
        }

        job.run(args, archive_args)

        self.assertIsInstance(job.archive_manager, ArchiveManager)
        result = job.result
        self.assertIsInstance(job.result, JobResult)
        self.assertEqual(result.job_id, '1234567890')
        self.assertEqual(result.job_number, 8)
        self.assertEqual(result.task_id, 'mytask')
        self.assertEqual(result.backend, 'git')
        self.assertEqual(result.category, 'commit')
        self.assertEqual(result.summary.last_uuid, '1375b60d3c23ac9b81da92523e4144abc4489d4c')
        self.assertEqual(result.summary.max_updated_on,
                         datetime.datetime(2014, 2, 12, 6, 10, 39, tzinfo=UTC))
        self.assertEqual(result.summary.last_updated_on,
                         datetime.datetime(2012, 8, 14, 17, 30, 13, tzinfo=UTC))
        self.assertEqual(result.summary.fetched, 9)
        self.assertEqual(result.summary.last_offset, None)

        commits = self.conn.lrange('items', 0, -1)
        commits = [pickle.loads(c) for c in commits]
        commits = [commit['data']['commit'] for commit in commits]

        expected = ['456a68ee1407a77f3e804a30dff245bb6c6b872f',
                    '51a3b654f252210572297f47597b31527c475fb8',
                    'ce8e0b86a1e9877f42fe9453ede418519115f367',
                    '589bb080f059834829a2a5955bebfd7c2baa110a',
                    'c6ba8f7a1058db3e6b4bc6f1090e932b107605fb',
                    'c0d66f92a95e31c77be08dc9d0f11a16715d1885',
                    '7debcf8a2f57f86663809c58b5c07a398be7674c',
                    '87783129c3f00d2c81a3a8e585eb86a47e39891a',
                    'bc57a9209f096a130dcc5ba7089a8663f758a703']

        self.assertEqual(commits, expected)

    def test_metadata(self):
        """Check if metadata parameters are correctly set"""

        job = PercevalJob('1234567890', 8, 'mytask',
                          'git', 'commit',
                          self.conn, 'items')
        args = {
            'uri': 'http://example.com/',
            'gitpath': os.path.join(self.dir, 'data/git_log.txt')
        }
        archive_args = {
            'archive_path': self.tmp_path,
            'fetch_from_archive': False
        }

        job.run(args, archive_args)

        items = self.conn.lrange('items', 0, -1)
        items = [pickle.loads(item) for item in items]

        for item in items:
            self.assertEqual(item['arthur_version'], __version__)
            self.assertEqual(item['job_id'], '1234567890')

    def test_run_not_found_parameters(self):
        """Check if it fails when a required backend parameter is not found"""

        job = PercevalJob('1234567890', 8, 'mytask',
                          'git', 'commit',
                          self.conn, 'items')
        args = {
            'uri': 'http://example.com/'
        }
        archive_args = {
            'archive_path': self.tmp_path,
            'fetch_from_archive': False
        }

        with self.assertRaises(AttributeError) as e:
            job.run(args, archive_args)
            self.assertEqual(e.exception.args[1], 'gitlog')

    @httpretty.activate
    def _test_fetch_from_archive(self):
        """Test run method fetching from the archive"""

        http_requests = setup_mock_bugzilla_server()

        expected = ['5a8a1e25dfda86b961b4146050883cbfc928f8ec',
                    '1fd4514e56f25a39ffd78eab19e77cfe4dfb7769',
                    '6a4cb244067c3cfa38f9f563e2ab3cd8ac21762f',
                    '7e033ed0110032ead6b918be43c1f3f88cd98fd7',
                    'f90d12b380ffdb47f2b6e96b321f08000181a9d6',
                    '4b166308f205121bc57704032acdc81b6c9bb8b1',
                    'b4009442d38f4241a4e22e3e61b7cd8ef5ced35c']

        # First, we fetch the bugs from the server, storing them
        # in a archive
        args = {
            'url': BUGZILLA_SERVER_URL,
            'max_bugs': 5
        }
        archive_args = {
            'archive_path': self.tmp_path,
            'fetch_from_archive': False
        }

        job = PercevalJob('1234567890', 8, 'mytask',
                          'bugzilla', 'issue',
                          self.conn, 'items')
        job.run(args, archive_args=archive_args)

        bugs = self.conn.lrange('items', 0, -1)
        bugs = [pickle.loads(b) for b in bugs]
        bugs = [bug['uuid'] for bug in bugs]
        self.conn.ltrim('items', 1, 0)

        result = job.result
        self.assertIsInstance(job.archive_manager, ArchiveManager)
        self.assertEqual(result.job_id, job.job_id)
        self.assertEqual(result.job_number, job.job_number)
        self.assertEqual(result.task_id, 'mytask')
        self.assertEqual(result.backend, 'bugzilla')
        self.assertEqual(result.category, 'issue')
        self.assertEqual(result.summary.last_uuid, 'b4009442d38f4241a4e22e3e61b7cd8ef5ced35c')
        self.assertEqual(result.summary.max_updated_on,
                         datetime.datetime(2011, 12, 8, 17, 58, 37, tzinfo=UTC))
        self.assertEqual(result.summary.last_updated_on,
                         datetime.datetime(2011, 12, 8, 17, 58, 37, tzinfo=UTC))
        self.assertEqual(result.summary.total, 7)
        self.assertEqual(result.summary.offset, None)

        self.assertEqual(len(http_requests), 13)
        self.assertListEqual(bugs, expected)

        # Now, we get the bugs from the archive.
        # The contents should be the same and there won't be
        # any new request to the server
        job_archive = PercevalJob('1234567890-bis', 'mytask', 'bugzilla', 'issue',
                                  self.conn, 'items')

        archive_args['fetch_from_archive'] = True
        job_archive.run(args, archive_args=archive_args)

        archived_bugs = self.conn.lrange('items', 0, -1)
        archived_bugs = [pickle.loads(b) for b in archived_bugs]
        archived_bugs = [bug['uuid'] for bug in archived_bugs]
        self.conn.ltrim('items', 1, 0)

        result = job_archive.result
        self.assertIsInstance(job.archive_manager, ArchiveManager)
        self.assertEqual(result.job_id, job_archive.job_id)
        self.assertEqual(result.task_id, 'mytask')
        self.assertEqual(result.backend, 'bugzilla')
        self.assertEqual(result.category, 'issue')
        self.assertEqual(result.summary.last_uuid, 'b4009442d38f4241a4e22e3e61b7cd8ef5ced35c')
        self.assertEqual(result.summary.max_updated_on,
                         datetime.datetime(2011, 12, 8, 17, 58, 37, tzinfo=UTC))
        self.assertEqual(result.summary.last_updated_on,
                         datetime.datetime(2011, 12, 8, 17, 58, 37, tzinfo=UTC))
        self.assertEqual(result.summary.total, 7)
        self.assertEqual(result.summary.offset, None)
        self.assertEqual(len(http_requests), 13)

        self.assertListEqual(archived_bugs, bugs)

    def test_initialize_archive_manager(self):
        """Test if the archive is initialized"""

        job = PercevalJob('1234567890', 8, 'mytask',
                          'git', 'commit',
                          self.conn, 'items')

        self.assertIsNone(job.archive_manager)

    def test_invalid_path_for_archive(self):
        """Test whether it raises an exception when the archive path is invalid"""

        job = PercevalJob('1234567890', 8, 'mytaks',
                          'git', 'commit',
                          self.conn, 'items')

        job.initialize_archive_manager(None)
        self.assertIsNone(job.archive_manager)

        with self.assertRaises(ValueError):
            job.initialize_archive_manager("")


class TestExecuteJob(TestBaseRQ):
    """Unit tests for execute_perceval_job"""

    def setUp(self):
        self.tmp_path = tempfile.mkdtemp(prefix='arthur_')
        super().setUp()

    def tearDown(self):
        shutil.rmtree(self.tmp_path)
        super().tearDown()

    def test_job(self):
        """Execute Git backend job"""

        backend_args = {
            'uri': 'http://example.com/',
            'gitpath': os.path.join(self.dir, 'data/git_log.txt')
        }
        archive_args = {}

        q = rq.Queue('queue', is_async=False)  # noqa: W606

        job = q.enqueue(execute_perceval_job,
                        backend='git', backend_args=backend_args, category='commit',
                        archive_args=archive_args,
                        qitems='items', task_id='mytask', job_number=8)

        result = job.return_value
        self.assertEqual(result.job_id, job.get_id())
        self.assertEqual(result.job_number, 8)
        self.assertEqual(result.task_id, 'mytask')
        self.assertEqual(result.backend, 'git')
        self.assertEqual(result.category, 'commit')
        self.assertEqual(result.summary.last_uuid, '1375b60d3c23ac9b81da92523e4144abc4489d4c')
        self.assertEqual(result.summary.max_updated_on,
                         datetime.datetime(2014, 2, 12, 6, 10, 39, tzinfo=UTC))
        self.assertEqual(result.summary.last_updated_on,
                         datetime.datetime(2012, 8, 14, 17, 30, 13, tzinfo=UTC))
        self.assertEqual(result.summary.total, 9)
        self.assertEqual(result.summary.max_offset, None)

        commits = self.conn.lrange('items', 0, -1)
        commits = [pickle.loads(c) for c in commits]
        commits = [(commit['job_id'], commit['data']['commit']) for commit in commits]

        expected = ['456a68ee1407a77f3e804a30dff245bb6c6b872f',
                    '51a3b654f252210572297f47597b31527c475fb8',
                    'ce8e0b86a1e9877f42fe9453ede418519115f367',
                    '589bb080f059834829a2a5955bebfd7c2baa110a',
                    'c6ba8f7a1058db3e6b4bc6f1090e932b107605fb',
                    'c0d66f92a95e31c77be08dc9d0f11a16715d1885',
                    '7debcf8a2f57f86663809c58b5c07a398be7674c',
                    '87783129c3f00d2c81a3a8e585eb86a47e39891a',
                    'bc57a9209f096a130dcc5ba7089a8663f758a703']

        for x in range(len(expected)):
            item = commits[x]
            self.assertEqual(item[0], result.job_id)
            self.assertEqual(item[1], expected[x])

    @httpretty.activate
    def test_failed_job(self):
        """Test if a failed job produce items an a partial result"""

        setup_mock_redmine_server(max_failures=2)

        backend_args = {
            'url': REDMINE_URL,
            'api_token': 'AAAA',
            'max_issues': 3
        }

        q = rq.Queue('queue', is_async=False)  # noqa: W606

        with self.assertRaises(requests.exceptions.HTTPError):
            _ = q.enqueue(execute_perceval_job,
                          job_id='test_failed_job',
                          backend='redmine', backend_args=backend_args,
                          category='issue',
                          qitems='items',
                          task_id='mytask', job_number=8)

        # The job failed but generated a partial result
        job = rq.job.Job.fetch('test_failed_job', connection=self.conn)

        result = job.meta['result']
        self.assertIsInstance(result, JobResult)
        self.assertEqual(result.job_id, job.get_id())
        self.assertEqual(result.job_number, 8)
        self.assertEqual(result.task_id, 'mytask')
        self.assertEqual(result.backend, 'redmine')
        self.assertEqual(result.summary.last_uuid, '3c3d67925b108a37f88cc6663f7f7dd493fa818c')
        self.assertEqual(result.summary.max_updated_on,
                         datetime.datetime(2011, 12, 8, 17, 58, 37, tzinfo=UTC))
        self.assertEqual(result.summary.last_updated_on,
                         datetime.datetime(2011, 12, 8, 17, 58, 37, tzinfo=UTC))
        self.assertEqual(result.summary.total, 3)
        self.assertEqual(result.summary.max_offset, None)

        issues = self.conn.lrange('items', 0, -1)
        issues = [pickle.loads(i) for i in issues]
        issues = [(issue['job_id'], issue['uuid']) for issue in issues]
        self.conn.ltrim('items', 1, 0)

        expected = ['91a8349c2f6ebffcccc49409529c61cfd3825563',
                    'c4aeb9e77fec8e4679caa23d4012e7cc36ae8b98',
                    '3c3d67925b108a37f88cc6663f7f7dd493fa818c']

        self.assertEqual(len(issues), len(expected))

        for x in range(len(expected)):
            item = issues[x]
            self.assertEqual(item[1], expected[x])

    def test_job_no_result(self):
        """Execute a Git backend job that will not produce any results"""

        backend_args = {
            'uri': 'http://example.com/',
            'gitpath': os.path.join(self.dir, 'data/git_log_empty.txt'),
            'from_date': datetime.datetime(2020, 1, 1, 1, 1, 1)
        }

        q = rq.Queue('queue', is_async=False)  # noqa: W606
        job = q.enqueue(execute_perceval_job,
                        backend='git', backend_args=backend_args,
                        category='commit',
                        qitems='items',
                        task_id='mytask', job_number=8)

        result = job.return_value
        self.assertEqual(result.job_id, job.get_id())
        self.assertEqual(result.job_number, 8)
        self.assertEqual(result.task_id, 'mytask')
        self.assertEqual(result.backend, 'git')
        self.assertEqual(result.category, 'commit')
        self.assertEqual(result.summary.last_uuid, None)
        self.assertEqual(result.summary.max_updated_on, None)
        self.assertEqual(result.summary.last_updated_on, None)
        self.assertEqual(result.summary.total, 0)
        self.assertEqual(result.summary.max_offset, None)

        commits = self.conn.lrange('items', 0, -1)
        commits = [pickle.loads(c) for c in commits]
        self.assertListEqual(commits, [])

    @httpretty.activate
    def test_job_archive(self):
        """Execute a Bugzilla backend job to fetch data from the archive"""

        after_dt = datetime_utcnow()
        http_requests = setup_mock_bugzilla_server()

        expected = ['5a8a1e25dfda86b961b4146050883cbfc928f8ec',
                    '1fd4514e56f25a39ffd78eab19e77cfe4dfb7769',
                    '6a4cb244067c3cfa38f9f563e2ab3cd8ac21762f',
                    '7e033ed0110032ead6b918be43c1f3f88cd98fd7',
                    'f90d12b380ffdb47f2b6e96b321f08000181a9d6',
                    '4b166308f205121bc57704032acdc81b6c9bb8b1',
                    'b4009442d38f4241a4e22e3e61b7cd8ef5ced35c']

        q = rq.Queue('queue', is_async=False)  # noqa: W606

        # First, we fetch the bugs from the server, storing them
        # in an archive
        backend_args = {
            'url': BUGZILLA_SERVER_URL,
            'max_bugs': 5
        }
        archive_args = {
            'archive_path': self.tmp_path,
            'fetch_from_archive': False,
            'archived_after': after_dt
        }

        job = q.enqueue(execute_perceval_job,
                        backend='bugzilla', backend_args=backend_args, category='bug',
                        qitems='items', task_id='mytask', job_number=8,
                        archive_args=archive_args)

        bugs = self.conn.lrange('items', 0, -1)
        bugs = [pickle.loads(b) for b in bugs]
        bugs = [bug['uuid'] for bug in bugs]
        self.conn.ltrim('items', 1, 0)

        result = job.return_value
        self.assertEqual(result.job_id, job.get_id())
        self.assertEqual(result.job_number, 8)
        self.assertEqual(result.task_id, 'mytask')
        self.assertEqual(result.backend, 'bugzilla')
        self.assertEqual(result.summary.last_uuid, 'b4009442d38f4241a4e22e3e61b7cd8ef5ced35c')
        self.assertEqual(result.summary.max_updated_on,
                         datetime.datetime(2015, 8, 12, 18, 32, 10, tzinfo=UTC))
        self.assertEqual(result.summary.last_updated_on,
                         datetime.datetime(2015, 8, 12, 18, 32, 10, tzinfo=UTC))
        self.assertEqual(result.summary.total, 7)

        self.assertEqual(len(http_requests), 13)
        self.assertListEqual(bugs, expected)

        # Now, we get the bugs from the archive.
        # The contents should be the same and there won't be
        # any new request to the server

        archive_args['fetch_from_archive'] = True
        job = q.enqueue(execute_perceval_job,
                        backend='bugzilla', backend_args=backend_args,
                        qitems='items', task_id='mytask', job_number=8,
                        category='bug',
                        archive_args=archive_args)

        archived_bugs = self.conn.lrange('items', 0, -1)
        archived_bugs = [pickle.loads(b) for b in archived_bugs]
        archived_bugs = [bug['uuid'] for bug in archived_bugs]
        self.conn.ltrim('items', 1, 0)

        result = job.return_value
        self.assertEqual(result.job_id, job.get_id())
        self.assertEqual(result.job_number, 8)
        self.assertEqual(result.task_id, 'mytask')
        self.assertEqual(result.backend, 'bugzilla')
        self.assertEqual(result.summary.last_uuid, 'b4009442d38f4241a4e22e3e61b7cd8ef5ced35c')
        self.assertEqual(result.summary.max_updated_on,
                         datetime.datetime(2015, 8, 12, 18, 32, 10, tzinfo=UTC))
        self.assertEqual(result.summary.last_updated_on,
                         datetime.datetime(2015, 8, 12, 18, 32, 10, tzinfo=UTC))
        self.assertEqual(result.summary.total, 7)

        self.assertEqual(len(http_requests), 13)

        self.assertListEqual(archived_bugs, bugs)

    def test_job_archiving_not_supported(self):
        """Check if it fails when archiving is not supported"""

        backend_args = {
            'uri': 'http://example.com/',
            'gitpath': os.path.join(self.dir, 'data/git_log.txt')
        }
        archive_args = {
            'archive_path': self.tmp_path,
            'fetch_from_archive': True
        }

        q = rq.Queue('queue', is_async=False)  # noqa: W606

        with self.assertRaises(AttributeError):
            _ = q.enqueue(execute_perceval_job,
                          backend='git', backend_args=backend_args,
                          category='commit',
                          qitems='items', task_id='mytask', job_number=8,
                          archive_args=archive_args)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#

import datetime
import unittest
import unittest.mock

import dateutil
from redis.exceptions import RedisError

from arthur.common import MAX_JOB_RETRIES, WAIT_FOR_QUEUING
from arthur.errors import (AlreadyExistsError,
                           NotFoundError,
                           TaskRegistryError)
from arthur.tasks import (ArchivingTaskConfig,
                          SchedulingTaskConfig,
                          Task,
                          JobData,
                          TaskRegistry,
                          TaskStatus,
                          logger)

from base import TestBaseRQ

INVALID_ARCHIVE_PATH_ERROR = "'archive_path' must be a str"
INVALID_ARCHIVED_AFTER_ERROR = "'archived_after' must be either a str or a datetime"
INVALID_ARCHIVED_AFTER_INVALID_DATE_ERROR = "is not a valid date"
INVALID_FETCH_FROM_ARCHIVE_ERROR = "'fetch_from_archive' must be a bool;"


class TestTask(unittest.TestCase):
    """Unit tests for Task"""

    @unittest.mock.patch('arthur.tasks.datetime_utcnow')
    def test_init(self, mock_utcnow):
        """Check arguments initialization"""

        mock_utcnow.return_value = datetime.datetime(2017, 1, 1,
                                                     tzinfo=dateutil.tz.tzutc())

        args = {
            'from_date': '1970-01-01',
            'component': 'test'
        }

        task = Task('mytask', 'git', 'commit', args)

        self.assertEqual(task.task_id, 'mytask')
        self.assertEqual(task.status, TaskStatus.NEW)
        self.assertEqual(task.age, 0)
        self.assertListEqual(task.jobs, [])
        self.assertEqual(task.backend, 'git')
        self.assertEqual(task.category, 'commit')
        self.assertDictEqual(task.backend_args, args)
        self.assertEqual(task.archiving_cfg, None)
        self.assertEqual(task.scheduling_cfg, None)
        self.assertEqual(task.created_on, 1483228800.0)

        # Test when archive and scheduler arguments are given
        archive = ArchivingTaskConfig('/tmp/archive', False,
                                      archived_after=None)
        sched = SchedulingTaskConfig(delay=10, max_retries=2,
                                     max_age=5)

        task = Task('mytask', 'git', 'commit', args,
                    archiving_cfg=archive, scheduling_cfg=sched)

        self.assertEqual(task.task_id, 'mytask')
        self.assertEqual(task.status, TaskStatus.NEW)
        self.assertEqual(task.age, 0)
        self.assertListEqual(task.jobs, [])
        self.assertEqual(task.backend, 'git')
        self.assertEqual(task.category, 'commit')
        self.assertDictEqual(task.backend_args, args)
        self.assertEqual(task.archiving_cfg, archive)
        self.assertEqual(task.scheduling_cfg, sched)
        self.assertEqual(task.created_on, 1483228800.0)

    def test_has_resuming(self):
        """Test if a task property returns it might be resumed or not"""

        task = Task('git_task', 'git', 'commit', {},
                    archiving_cfg=None, scheduling_cfg=None)
        self.assertEqual(task.has_resuming(), True)

        task = Task('gerrit_task', 'gerrit', 'review', {},
                    archiving_cfg=None, scheduling_cfg=None)
        self.assertEqual(task.has_resuming(), False)

    def test_backend_not_found(self):
        """Test if it raises an exception when a backend is not found"""

        archive = ArchivingTaskConfig('/tmp/archive', False,
                                      archived_after=None)
        sched = SchedulingTaskConfig(delay=10, max_retries=2,
                                     max_age=5)

        with self.assertRaises(NotFoundError) as e:
            _ = Task('mytask', 'mock_backend', 'mock_item', {},
                     archiving_cfg=archive, scheduling_cfg=sched)
            self.assertEqual(e.exception.element, 'mock_backend')

    def test_set_job(self):
        """Check whether the job is added to the task"""

        args = {
            'from_date': '1970-01-01',
            'component': 'test'
        }
        category = 'commit'

        task = Task('mytask', 'git', category, args)

        task.set_job('job-0', 0)
        task.set_job('job-1', 1)

        expected_job_0 = JobData('job-0', 0)
        expected_job_1 = JobData('job-1', 1)

        self.assertEqual(len(task.jobs), 2)
        self.assertListEqual(task.jobs, [expected_job_0, expected_job_1])

    def test_to_dict(self):
        """Check whether the object is converted into a dict"""

        args = {
            'from_date': '1970-01-01',
            'component': 'test'
        }
        category = 'commit'
        archive = ArchivingTaskConfig('/tmp/archive', False,
                                      archived_after=None)
        sched = SchedulingTaskConfig(delay=10, max_retries=2,
                                     max_age=5, queue='myqueue')
        before = datetime.datetime.now().timestamp()

        task = Task('mytask', 'git', category, args,
                    archiving_cfg=archive, scheduling_cfg=sched)
        task.set_job('job-0', 0)
        task.set_job('job-1', 1)

        d = task.to_dict()

        expected_job_0 = {
            'job_id': 'job-0',
            'job_number': 0
        }
        expected_job_1 = {
            'job_id': 'job-1',
            'job_number': 1
        }
        expected = {
            'task_id': 'mytask',
            'status': 'NEW',
            'age': 0,
            'num_failures': 0,
            'jobs': [expected_job_0, expected_job_1],
            'backend': 'git',
            'backend_args': args,
            'category': category,
            'archiving_cfg': {
                'archive_path': '/tmp/archive',
                'archived_after': None,
                'fetch_from_archive': False
            },
            'scheduling_cfg': {
                'delay': 10,
                'max_retries': 2,
                'max_age': 5,
                'queue': 'myqueue'
            }
        }

        created_on = d.pop('created_on')
        self.assertGreater(created_on, before)
        self.assertDictEqual(d, expected)


class TestTaskRegistry(TestBaseRQ):
    """Unit tests for TaskRegistry"""

    def test_empty_registry(self):
        """Check empty registry on init"""

        registry = TaskRegistry(self.conn)
        tasks = registry.tasks

        self.assertListEqual(tasks, [])

    def test_tasks(self):
        """Test to list tasks in the registry"""

        registry = TaskRegistry(self.conn)
        expected = []
        for i in range(0, 20):
            task_id = 'mytask{}'.format(i)
            expected.append(task_id)
            registry.add(task_id, 'git', 'commit', {})

        tasks = registry.tasks
        self.assertEqual(len(tasks), 20)

        for t in tasks:
            self.assertIn(t.task_id, expected)

    @unittest.mock.patch('redis.StrictRedis.scan')
    def test_tasks_error(self, mock_redis_set):
        """Test whether the lock is released in case of a Redis error and a TaskRegistryError is thrown"""

        mock_redis_set.side_effect = RedisError

        registry = TaskRegistry(self.conn)
        registry.add('mytask', 'git', 'commit', {})
        self.assertGreater(registry._rwlock._readers_mutex._value, 0)
        with self.assertRaises(TaskRegistryError):
            _ = registry.tasks
            self.assertGreater(registry._rwlock._readers_mutex._value, 0)

    def test_matched_tasks(self):
        """Test to list tasks in the registry when other data is stored in Redis"""

        registry = TaskRegistry(self.conn)

        self.conn.set("key1", 'value')
        self.conn.set("key2", 1)
        self.conn.set("key3", b'x')

        expected = []
        for i in range(0, 20):
            task_id = 'mytask{}'.format(i)
            expected.append(task_id)
            registry.add(task_id, 'git', 'commit', {})

        tasks = registry.tasks
        self.assertEqual(len(tasks), 20)

        for t in tasks:
            self.assertIn(t.task_id, expected)

    def test_add_task(self):
        """Test to add tasks to the registry"""

        args = {
            'from_date': '1970-01-01',
            'component': 'test'
        }
        archive = ArchivingTaskConfig('/tmp/archive', False,
                                      archived_after=None)
        sched = SchedulingTaskConfig(delay=10, max_retries=2)

        registry = TaskRegistry(self.conn)
        before = datetime.datetime.now().timestamp()
        _ = registry.add('mytask', 'git', 'commit', args)

        tasks = registry.tasks
        self.assertEqual(len(tasks), 1)

        task = tasks[0]
        self.assertIsInstance(task, Task)
        self.assertEqual(task.task_id, 'mytask')
        self.assertEqual(task.status, TaskStatus.NEW)
        self.assertEqual(task.age, 0)
        self.assertListEqual(task.jobs, [])
        self.assertEqual(task.category, 'commit')
        self.assertEqual(task.backend, 'git')
        self.assertDictEqual(task.backend_args, args)
        self.assertEqual(task.archiving_cfg, None)
        self.assertEqual(task.scheduling_cfg, None)
        self.assertGreater(task.created_on, before)

        before = datetime.datetime.now().timestamp()
        _ = registry.add('atask', 'git', 'commit', args,
                         archiving_cfg=archive, scheduling_cfg=sched)

        tasks = registry.tasks
        self.assertEqual(len(tasks), 2)

        task0 = tasks[0]
        self.assertIsInstance(task0, Task)
        self.assertEqual(task0.task_id, 'atask')
        self.assertEqual(task0.status, TaskStatus.NEW)
        self.assertEqual(task.age, 0)
        self.assertListEqual(task.jobs, [])
        self.assertEqual(task0.backend, 'git')
        self.assertEqual(task0.category, 'commit')
        self.assertDictEqual(task0.backend_args, args)
        self.assertEqual(task0.archiving_cfg.archive_path, archive.archive_path)
        self.assertEqual(task0.archiving_cfg.archived_after, archive.archived_after)
        self.assertEqual(task0.archiving_cfg.fetch_from_archive, archive.fetch_from_archive)
        self.assertEqual(task0.scheduling_cfg.delay, sched.delay)
        self.assertEqual(task0.scheduling_cfg.max_retries, sched.max_retries)
        self.assertGreater(task0.created_on, before)

        task1 = tasks[1]
        self.assertEqual(task1.task_id, 'mytask')
        self.assertGreater(task0.created_on, task1.created_on)

    def test_add_existing_task(self):
        """Check if it raises an exception when an existing task is added again"""

        registry = TaskRegistry(self.conn)
        registry.add('mytask', 'git', 'commit', {})

        with self.assertRaises(AlreadyExistsError):
            registry.add('mytask', 'new_backend', 'commit', {})

        # Only one tasks is on the registry
        tasks = registry.tasks
        self.assertEqual(len(tasks), 1)

        task = tasks[0]
        self.assertEqual(task.task_id, 'mytask')
        self.assertEqual(task.backend, 'git')
        self.assertEqual(task.category, 'commit')
        self.assertDictEqual(task.backend_args, {})

    @unittest.mock.patch('redis.StrictRedis.set')
    def test_add_task_error(self, mock_redis_cmd):
        """Test whether the lock is released in case of a Redis error and a TaskRegistryError is thrown"""

        mock_redis_cmd.side_effect = RedisError

        registry = TaskRegistry(self.conn)
        self.assertGreater(registry._rwlock._access_mutex._value, 0)
        with self.assertRaises(TaskRegistryError):
            registry.add('mytask', 'git', 'commit', {})
            self.assertGreater(registry._rwlock._access_mutex._value, 0)

    def test_remove_task(self):
        """Test to remove a task from the registry"""

        args = {
            'from_date': '1970-01-01',
            'component': 'test'
        }

        registry = TaskRegistry(self.conn)
        registry.add('mytask', 'git', 'commit', args)
        registry.add('newtask', 'bugzilla', 'issue', None)
        registry.add('atask', 'git', 'commit', None)

        tasks = registry.tasks
        self.assertEqual(len(tasks), 3)

        registry.remove('newtask')

        tasks = registry.tasks
        self.assertEqual(len(tasks), 2)
        self.assertEqual(tasks[0].task_id, 'atask')
        self.assertEqual(tasks[1].task_id, 'mytask')

    def test_remove_not_found(self):
        """Check whether it raises an exception when a task is not found"""

        registry = TaskRegistry(self.conn)

        with self.assertRaises(NotFoundError):
            registry.remove('mytask')

        registry.add('task', 'git', "mock_category", {})

        with self.assertRaises(NotFoundError):
            registry.remove('mytask')

        self.assertEqual(len(registry.tasks), 1)

    @unittest.mock.patch('redis.StrictRedis.delete')
    def test_remove_task_error(self, mock_redis_cmd):
        """Test whether the lock is released in case of a Redis error and an error TaskRegistryError is thrown"""

        mock_redis_cmd.side_effect = RedisError

        registry = TaskRegistry(self.conn)
        registry.add('mytask', 'git', 'commit', {})
        self.assertGreater(registry._rwlock._access_mutex._value, 0)
        with self.assertRaises(TaskRegistryError):
            registry.remove('mytask')
            self.assertGreater(registry._rwlock._access_mutex._value, 0)

    def test_get_task(self):
        """Test to get a task from the registry"""

        args = {
            'from_date': '1970-01-01',
            'component': 'test'
        }

        registry = TaskRegistry(self.conn)
        registry.add('mytask', 'git', 'commit', args)
        registry.add('newtask', 'bugzilla', 'issue', None)
        registry.add('atask', 'git', 'commit', None)

        task = registry.get('atask')
        self.assertIsInstance(task, Task)
        self.assertEqual(task.task_id, 'atask')
        self.assertEqual(task.status, TaskStatus.NEW)
        self.assertEqual(task.category, 'commit')
        self.assertEqual(task.backend, 'git')
        self.assertEqual(task.backend_args, None)

    def test_get_task_not_found(self):
        """Check whether it raises an exception when a task is not found"""

        registry = TaskRegistry(self.conn)

        with self.assertRaises(NotFoundError):
            registry.get('mytask')

        registry.add('newtask', 'git', 'mocked_category', {})

        with self.assertRaises(NotFoundError):
            registry.get('mytask')

    @unittest.mock.patch('redis.StrictRedis.get')
    def test_get_task_error(self, mock_redis_cmd):
        """Test whether the lock is released in case of a Redis error and a TaskRegistryError is thrown"""

        mock_redis_cmd.side_effect = RedisError

        registry = TaskRegistry(self.conn)
        registry.add('mytask', 'git', 'commit', {})
        self.assertGreater(registry._rwlock._access_mutex._value, 0)
        with self.assertRaises(TaskRegistryError):
            registry.get('mytask')
            self.assertGreater(registry._rwlock._access_mutex._value, 0)

    def test_update_task(self):
        """Test to update tasks in the registry"""

        args = {
            'from_date': '1970-01-01',
            'component': 'test'
        }

        registry = TaskRegistry(self.conn)
        before = datetime.datetime.now().timestamp()
        _ = registry.add('mytask', 'git', 'commit', args)

        tasks = registry.tasks
        self.assertEqual(len(tasks), 1)

        task = tasks[0]
        self.assertIsInstance(task, Task)
        self.assertEqual(task.task_id, 'mytask')
        self.assertEqual(task.status, TaskStatus.NEW)
        self.assertEqual(task.age, 0)
        self.assertListEqual(task.jobs, [])
        self.assertEqual(task.category, 'commit')
        self.assertEqual(task.backend, 'git')
        self.assertDictEqual(task.backend_args, args)
        self.assertEqual(task.archiving_cfg, None)
        self.assertEqual(task.scheduling_cfg, None)
        self.assertGreater(task.created_on, before)

        task.status = TaskStatus.COMPLETED
        task.age = 1
        registry.update('mytask', task)

        tasks = registry.tasks
        self.assertEqual(len(tasks), 1)

        task = tasks[0]
        self.assertIsInstance(task, Task)
        self.assertEqual(task.task_id, 'mytask')
        self.assertEqual(task.status, TaskStatus.COMPLETED)
        self.assertEqual(task.age, 1)
        self.assertListEqual(task.jobs, [])
        self.assertEqual(task.category, 'commit')
        self.assertEqual(task.backend, 'git')
        self.assertDictEqual(task.backend_args, args)
        self.assertEqual(task.archiving_cfg, None)
        self.assertEqual(task.scheduling_cfg, None)
        self.assertGreater(task.created_on, before)

    def test_update_task_not_found(self):
        """Check whether it raises an exception when a task is not found"""

        registry = TaskRegistry(self.conn)

        with self.assertLogs(logger, level='WARNING') as cm:
            registry.update('mytask', None)
            self.assertEqual(cm.output[0], 'WARNING:arthur.tasks:Task mytask not found, adding it')

    @unittest.mock.patch('redis.StrictRedis.exists')
    def test_update_task_error(self, mock_redis_cmd):
        """Test whether the lock is released in case of a Redis error and a TaskRegistryError is thrown"""

        mock_redis_cmd.side_effect = RedisError

        registry = TaskRegistry(self.conn)
        self.assertGreater(registry._rwlock._access_mutex._value, 0)
        with self.assertRaises(TaskRegistryError):
            registry.update('mytask', None)
            self.assertGreater(registry._rwlock._access_mutex._value, 0)


class TestArchivingTaskConfig(unittest.TestCase):
    """Unit tests for ArchivingTaskConfig"""

    def test_init(self):
        """Test whether object properties are initialized"""

        dt = datetime.datetime(2001, 12, 1, 23, 15, 32,
                               tzinfo=dateutil.tz.tzutc())

        archiving_cfg = ArchivingTaskConfig('/tmp/archive', True,
                                            archived_after=dt)
        self.assertEqual(archiving_cfg.archive_path, '/tmp/archive')
        self.assertEqual(archiving_cfg.fetch_from_archive, True)
        self.assertEqual(archiving_cfg.archived_after, dt)

    def test_set_archive_path(self):
        """Test if archive_path property can be set"""

        archiving_cfg = ArchivingTaskConfig('/tmp/archive', False)
        self.assertEqual(archiving_cfg.archive_path, '/tmp/archive')

        archiving_cfg.archive_path = '/tmp/oldarchive'
        self.assertEqual(archiving_cfg.archive_path, '/tmp/oldarchive')

    def test_set_invalid_archive_path(self):
        """Check if an exception is raised for invalid archive_path values"""

        with self.assertRaisesRegex(ValueError, INVALID_ARCHIVE_PATH_ERROR):
            archiving_cfg = ArchivingTaskConfig(5.0, False)

        archiving_cfg = ArchivingTaskConfig('/tmp/archive', False)

        with self.assertRaisesRegex(ValueError, INVALID_ARCHIVE_PATH_ERROR):
            archiving_cfg.archive_path = 1.0

        with self.assertRaisesRegex(ValueError, INVALID_ARCHIVE_PATH_ERROR):
            archiving_cfg.archive_path = True

        self.assertEqual(archiving_cfg.archive_path, '/tmp/archive')

    def test_set_fetch_from_archive(self):
        """Test if fetch_from_archive property can be set"""

        archiving_cfg = ArchivingTaskConfig('/tmp/archive', True)
        self.assertEqual(archiving_cfg.fetch_from_archive, True)

        archiving_cfg.fetch_from_archive = False
        self.assertEqual(archiving_cfg.fetch_from_archive, False)

    def test_set_invalid_fetch_from_archive(self):
        """Check if an exception is raised for invalid fetch_from_archive values"""

        with self.assertRaisesRegex(ValueError, INVALID_FETCH_FROM_ARCHIVE_ERROR):
            archiving_cfg = ArchivingTaskConfig('/tmp/archive', 'False')

        archiving_cfg = ArchivingTaskConfig('/tmp/archive', True)

        with self.assertRaisesRegex(ValueError, INVALID_FETCH_FROM_ARCHIVE_ERROR):
            archiving_cfg.fetch_from_archive = 1.0

        with self.assertRaisesRegex(ValueError, INVALID_FETCH_FROM_ARCHIVE_ERROR):
            archiving_cfg.fetch_from_archive = 'False'

        self.assertEqual(archiving_cfg.fetch_from_archive, True)

    def test_set_archived_after(self):
        """Test if archived_after property can be set"""

        archiving_cfg = ArchivingTaskConfig('/tmp/archive', True)
        self.assertEqual(archiving_cfg.archived_after, None)

        dt = datetime.datetime(2001, 12, 1, 23, 15, 32,
                               tzinfo=dateutil.tz.tzutc())

        archiving_cfg = ArchivingTaskConfig('/tmp/archive', True,
                                            archived_after=dt)
        self.assertEqual(archiving_cfg.archived_after, dt)

        dt = datetime.datetime(2018, 1, 1,
                               tzinfo=dateutil.tz.tzutc())

        archiving_cfg.archived_after = dt
        self.assertEqual(archiving_cfg.archived_after, dt)

    def test_set_archived_after_to_utc(self):
        """Check whether dates are converted to UTC"""

        dt = datetime.datetime(2001, 12, 1, 23, 15, 32,
                               tzinfo=dateutil.tz.tzoffset(None, -21600))
        expected = datetime.datetime(2001, 12, 2, 5, 15, 32,
                                     tzinfo=dateutil.tz.tzutc())

        archiving_cfg = ArchivingTaskConfig('/tmp/archive', True,
                                            archived_after=dt)
        # Date should have been converted to UTC
        self.assertEqual(archiving_cfg.archived_after, expected)

        archiving_cfg.archived_after = datetime.datetime(2001, 12, 2, 5, 15, 32)
        self.assertEqual(archiving_cfg.archived_after, expected)

    def test_set_archived_after_from_date_string(self):
        """Test if archived_after property can be set from a date string"""

        date_str = '2001-12-01 23:15:32 -0600'

        expected = datetime.datetime(2001, 12, 2, 5, 15, 32,
                                     tzinfo=dateutil.tz.tzutc())

        archiving_cfg = ArchivingTaskConfig('/tmp/archive', True,
                                            archived_after=date_str)
        self.assertEqual(archiving_cfg.archived_after, expected)

    def test_set_invalid_archived_after(self):
        """Check if an exception is raised for invalid archived_after values"""

        with self.assertRaisesRegex(ValueError, INVALID_ARCHIVED_AFTER_ERROR):
            archiving_cfg = ArchivingTaskConfig('/tmp/archive', False,
                                                archived_after=1.0)
        with self.assertRaisesRegex(ValueError, INVALID_ARCHIVED_AFTER_INVALID_DATE_ERROR):
            archiving_cfg = ArchivingTaskConfig('/tmp/archive', False,
                                                archived_after='this date')

        archiving_cfg = ArchivingTaskConfig('/tmp/archive', True)

        with self.assertRaisesRegex(ValueError, INVALID_ARCHIVED_AFTER_ERROR):
            archiving_cfg.archived_after = 1.0

        with self.assertRaisesRegex(ValueError, INVALID_ARCHIVED_AFTER_INVALID_DATE_ERROR):
            archiving_cfg.archived_after = ''

        self.assertEqual(archiving_cfg.archived_after, None)


class TestSchedulingTaskConfig(unittest.TestCase):
    """Unit tests for TaskRegistry"""

    def test_init(self):
        """Test whether object properties are initialized"""

        scheduling_cfg = SchedulingTaskConfig()
        self.assertEqual(scheduling_cfg.delay, WAIT_FOR_QUEUING)
        self.assertEqual(scheduling_cfg.max_retries, MAX_JOB_RETRIES)
        self.assertEqual(scheduling_cfg.max_age, None)
        self.assertEqual(scheduling_cfg.queue, None)

        scheduling_cfg = SchedulingTaskConfig(delay=5, max_retries=1,
                                              max_age=10, queue='myqueue')
        self.assertEqual(scheduling_cfg.delay, 5)
        self.assertEqual(scheduling_cfg.max_retries, 1)
        self.assertEqual(scheduling_cfg.max_age, 10)
        self.assertEqual(scheduling_cfg.queue, 'myqueue')

    def test_set_delay(self):
        """Test if delay property can be set"""

        scheduling_cfg = SchedulingTaskConfig(delay=5)
        self.assertEqual(scheduling_cfg.delay, 5)

        scheduling_cfg.delay = 1
        self.assertEqual(scheduling_cfg.delay, 1)

    def test_set_invalid_delay(self):
        """Check if an exception is raised for invalid delay values"""

        with self.assertRaises(ValueError):
            scheduling_cfg = SchedulingTaskConfig(delay=5.0)

        scheduling_cfg = SchedulingTaskConfig(delay=5)

        with self.assertRaises(ValueError):
            scheduling_cfg.delay = 1.0

        with self.assertRaises(ValueError):
            scheduling_cfg.delay = '1'

        self.assertEqual(scheduling_cfg.delay, 5)

    def test_set_max_retries(self):
        """Test if max_retries property can be set"""

        scheduling_cfg = SchedulingTaskConfig(max_retries=3)
        self.assertEqual(scheduling_cfg.max_retries, 3)

        scheduling_cfg.max_retries = 1
        self.assertEqual(scheduling_cfg.max_retries, 1)

    def test_set_invalid_max_retries(self):
        """Check if an exception is raised for invalid max_retries values"""

        with self.assertRaises(ValueError):
            scheduling_cfg = SchedulingTaskConfig(max_retries=2.0)

        scheduling_cfg = SchedulingTaskConfig(max_retries=3)

        with self.assertRaises(ValueError):
            scheduling_cfg.max_retries = 5.0

        with self.assertRaises(ValueError):
            scheduling_cfg.max_retries = '5'

        self.assertEqual(scheduling_cfg.max_retries, 3)

    def test_set_max_age(self):
        """Test if max_age property can be set"""

        scheduling_cfg = SchedulingTaskConfig(max_age=3)
        self.assertEqual(scheduling_cfg.max_age, 3)

        scheduling_cfg.max_age = None
        self.assertEqual(scheduling_cfg.max_age, None)

    def test_set_invalid_max_age(self):
        """Check if an exception is raised for invalid max_age values"""

        with self.assertRaises(ValueError):
            _ = SchedulingTaskConfig(max_age=2.0)

        scheduling_cfg = SchedulingTaskConfig(max_age=3)

        with self.assertRaises(ValueError):
            scheduling_cfg.max_age = 0

        with self.assertRaises(ValueError):
            scheduling_cfg.max_age = -1

        with self.assertRaises(ValueError):
            scheduling_cfg.max_age = '5'

        self.assertEqual(scheduling_cfg.max_age, 3)

    def test_set_queue(self):
        """Test if queue property can be set"""

        scheduling_cfg = SchedulingTaskConfig(queue='myqueue')
        self.assertEqual(scheduling_cfg.queue, 'myqueue')

        scheduling_cfg.queue = None
        self.assertEqual(scheduling_cfg.queue, None)

    def test_set_invalid_queue(self):
        """Check if an exception is raised for invalid queue values"""

        with self.assertRaises(ValueError):
            _ = SchedulingTaskConfig(queue=1.0)

        scheduling_cfg = SchedulingTaskConfig(queue='myqueue')

        with self.assertRaises(ValueError):
            scheduling_cfg.queue = 0

        with self.assertRaises(ValueError):
            scheduling_cfg.queue = -1

        self.assertEqual(scheduling_cfg.queue, 'myqueue')

    def test_from_dict(self):
        """Check if an object is created when its properties are given from a dict"""

        opts = {
            'delay': 1,
            'max_retries': 3,
            'max_age': 5,
            'queue': 'myqueue'
        }

        scheduling_cfg = SchedulingTaskConfig.from_dict(opts)
        self.assertIsInstance(scheduling_cfg, SchedulingTaskConfig)
        self.assertEqual(scheduling_cfg.delay, 1)
        self.assertEqual(scheduling_cfg.max_retries, 3)
        self.assertEqual(scheduling_cfg.max_age, 5)
        self.assertEqual(scheduling_cfg.queue, 'myqueue')

    def test_from_dict_max_age_none(self):
        """Check if an object is created from a dict when max_age is None"""

        # Test None options
        opts = {
            'max_age': None
        }

        scheduling_cfg = SchedulingTaskConfig.from_dict(opts)
        self.assertIsInstance(scheduling_cfg, SchedulingTaskConfig)
        self.assertEqual(scheduling_cfg.max_age, None)

    def test_from_dict_queue_none(self):
        """Check if an object is created from a dict when queue is None"""

        # Test None options
        opts = {
            'queue': None
        }

        scheduling_cfg = SchedulingTaskConfig.from_dict(opts)
        self.assertIsInstance(scheduling_cfg, SchedulingTaskConfig)
        self.assertEqual(scheduling_cfg.queue, None)

    def test_from_dict_invalid_option(self):
        """Check if an exception is raised when an invalid option is given"""

        opts = {
            'delay': 1,
            'max_retries': 3,
            'max_age': None,
            'custom_param': 'myparam'
        }

        with self.assertRaisesRegex(ValueError,
                                    "unknown 'custom_param' task config parameter"):
            _ = SchedulingTaskConfig.from_dict(opts)

    def test_to_dict(self):
        """Check whether the object is converted into a dictionary"""

        scheduling_cfg = SchedulingTaskConfig(delay=5, max_retries=1,
                                              max_age=5, queue='myqueue')
        d = scheduling_cfg.to_dict()

        expected = {
            'delay': 5,
            'max_retries': 1,
            'max_age': 5,
            'queue': 'myqueue'
        }

        self.assertDictEqual(d, expected)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#

import unittest

import fakeredis

from grimoirelab_toolkit.datetime import datetime_utcnow

from arthur.common import CH_PUBSUB
from arthur.events import (JobEventType,
                           JobEvent,
                           JobEventsListener)

from base import TestBaseRQ


class MockJobResult:
    """Class to test/mock job results"""

    def __init__(self, result, category):
        self.result = result
        self.category = category
        self.timestamp = datetime_utcnow()


class TestJobEvent(unittest.TestCase):
    """Unit tests for JobEvent class"""

    def test_initialization(self):
        """Test if the instance is correctly initialized"""

        dt_before = datetime_utcnow()
        event = JobEvent(JobEventType.COMPLETED, '1', 'mytask', None)
        dt_after = datetime_utcnow()

        self.assertEqual(event.type, JobEventType.COMPLETED)
        self.assertEqual(event.job_id, '1')
        self.assertEqual(event.task_id, 'mytask')
        self.assertEqual(event.payload, None)
        self.assertGreater(event.timestamp, dt_before)
        self.assertLess(event.timestamp, dt_after)

        dt_before = datetime_utcnow()
        event = JobEvent(JobEventType.FAILURE, '2', 'myothertask', "Error")
        dt_after = datetime_utcnow()

        self.assertEqual(event.type, JobEventType.FAILURE)
        self.assertEqual(event.job_id, '2')
        self.assertEqual(event.task_id, 'myothertask')
        self.assertEqual(event.payload, "Error")
        self.assertGreater(event.timestamp, dt_before)
        self.assertLess(event.timestamp, dt_after)

    def test_unique_identifier(self):
        """Test if different identifiers create unique identifiers"""

        event_a = JobEvent(JobEventType.COMPLETED, '1', 'A', None)
        event_b = JobEvent(JobEventType.COMPLETED, '2', 'B', None)
        event_c = JobEvent(JobEventType.FAILURE, '3', 'C', None)

        self.assertNotEqual(event_a.uuid, None)
        self.assertNotEqual(event_a.uuid, event_b.uuid)
        self.assertNotEqual(event_b.uuid, event_c.uuid)
        self.assertNotEqual(event_c.uuid, event_a.uuid)

    def test_serializer(self):
        """Test if an event is properly serialized and deserialized"""

        result = MockJobResult(10, 'mockbackend')
        event_a = JobEvent(JobEventType.COMPLETED, '1', 'A', result)

        data = event_a.serialize()
        event = JobEvent.deserialize(data)

        self.assertIsInstance(event, JobEvent)
        self.assertEqual(event.uuid, event_a.uuid)
        self.assertEqual(event.timestamp, event_a.timestamp)
        self.assertEqual(event.type, event_a.type)
        self.assertEqual(event.job_id, event_a.job_id)
        self.assertEqual(event.task_id, event_a.task_id)

        payload = event.payload
        self.assertIsInstance(payload, MockJobResult)
        self.assertEqual(payload.result, result.result)
        self.assertEqual(payload.category, result.category)
        self.assertEqual(payload.timestamp, result.timestamp)


class MockPubSub:
    """Mock class to have a non-blocking job events listener"""

    def __init__(self, fake_events):
        self.fake_events = fake_events

    def subscribe(self, *args, **kwargs):
        pass

    def listen(self):
        for event in self.fake_events:
            yield {
                'type': 'message',
                'data': event.serialize()
            }


class MockRedisPubSubConnection(fakeredis.FakeStrictRedis):
    """Class to produce mock pubsub objects"""

    def __init__(self, fake_events, *args, **kwargs):
        self.fake_events = fake_events
        super().__init__(*args, **kwargs)

    def pubsub(self):
        return MockPubSub(self.fake_events)


class TestJobEventsListener(TestBaseRQ):
    """Unit tests for JobEventsListener class"""

    def test_initialization(self):
        """Test if the instance is correctly initialized"""

        listener = JobEventsListener(self.conn)
        self.assertEqual(listener.conn, self.conn)
        self.assertEqual(listener.events_channel, CH_PUBSUB)
        self.assertDictEqual(listener.handlers, {})

        listener = JobEventsListener(self.conn,
                                     events_channel='events')

        self.assertEqual(listener.conn, self.conn)
        self.assertEqual(listener.events_channel, 'events')
        self.assertDictEqual(listener.handlers, {})

    def test_subscribe(self):
        """Test if a listener is subscribed to a set of events"""

        def handle_successful_job(job):
            pass

        def handle_failed_job(job):
            pass

        listener = JobEventsListener(self.conn)

        listener.subscribe(JobEventType.COMPLETED, handle_successful_job)
        listener.subscribe(JobEventType.FAILURE, handle_failed_job)

        self.assertEqual(listener.handlers[JobEventType.COMPLETED],
                         handle_successful_job)
        self.assertEqual(listener.handlers[JobEventType.FAILURE],
                         handle_failed_job)

    def test_replace_subscription(self):
        """Test if a handler is replaced when subscription is called multiple times"""

        def handle_generic_event(job):
            pass

        listener = JobEventsListener(self.conn)

        listener.subscribe(JobEventType.COMPLETED, handle_generic_event)
        self.assertEqual(listener.handlers[JobEventType.COMPLETED],
                         handle_generic_event)

        def handle_specific_event(job):
            pass

        # This should replace the handler by the newer one
        listener.subscribe(JobEventType.COMPLETED, handle_specific_event)
        self.assertEqual(listener.handlers[JobEventType.COMPLETED],
                         handle_specific_event)

    def test_subscribe_invalid_type(self):
        """Check if an exception is raised when the event type is invalid"""

        def handle_generic_event(job):
            pass

        listener = JobEventsListener(self.conn)

        with self.assertRaisesRegex(TypeError, "'str' object is not a JobEventType"):
            listener.subscribe('COMPLETED', handle_generic_event)

    def test_unsubscribe(self):
        """Check if a listener removes its subscription from a set of events"""

        def handle_generic_event(job):
            pass

        listener = JobEventsListener(self.conn)

        listener.subscribe(JobEventType.COMPLETED, handle_generic_event)
        self.assertEqual(listener.handlers[JobEventType.COMPLETED],
                         handle_generic_event)

        listener.unsubscribe(JobEventType.COMPLETED)
        self.assertEqual(listener.handlers[JobEventType.COMPLETED], None)

    def test_unsubscribe_invalid_type(self):
        """Check if an exception is raised when the event type is invalid"""

        listener = JobEventsListener(self.conn)

        with self.assertRaisesRegex(TypeError, "'str' object is not a JobEventType"):
            listener.unsubscribe('COMPLETED')

    def test_listen(self):
        """Check if if listens and handles event messages.

        Due to listening on a channel is a blocking call (because
        Redis lib implementation), this test mocks the connection
        with it to simulate events reception.
        """

        class TrackedEvents:
            def __init__(self):
                self.handled = 0
                self.ok = 0
                self.failures = 0

        global tracked_events
        tracked_events = TrackedEvents()

        def handle_successful_job(job):
            tracked_events.handled += 1
            tracked_events.ok += 1

        def handle_failed_job(job):
            tracked_events.handled += 1
            tracked_events.failures += 1

        events = [
            JobEvent(JobEventType.COMPLETED, 1, 'A', MockJobResult(20, 'A')),
            JobEvent(JobEventType.FAILURE, 2, 'B', 'ERROR'),
            JobEvent(JobEventType.FAILURE, 3, 'C', 'ERROR'),
            JobEvent(JobEventType.FAILURE, 4, 'D', 'ERROR'),
            JobEvent(JobEventType.COMPLETED, 5, 'E', MockJobResult(22, 'B')),
            JobEvent(JobEventType.FAILURE, 6, 'F', 'ERROR'),
            JobEvent(JobEventType.UNDEFINED, 7, 'G', 'OK'),
        ]

        conn = MockRedisPubSubConnection(events)
        listener = JobEventsListener(conn)
        listener.subscribe(JobEventType.COMPLETED, handle_successful_job)
        listener.subscribe(JobEventType.FAILURE, handle_failed_job)

        listener.run()

        # UNDEFINED event was not handled
        self.assertEqual(tracked_events.handled, 6)
        self.assertEqual(tracked_events.ok, 2)
        self.assertEqual(tracked_events.failures, 4)

    def test_listen_after_unsubscribing(self):
        """Check if only handles subscribed messages after unsubscribing.

        Due to listening on a channel is a blocking call (because
        Redis lib implementation), this test mocks the connection
        with it to simulate events reception.
        """

        class TrackedEvents:
            def __init__(self):
                self.handled = 0
                self.ok = 0
                self.failures = 0

        global tracked_events
        tracked_events = TrackedEvents()

        def handle_successful_job(job):
            tracked_events.handled += 1
            tracked_events.ok += 1

        def handle_failed_job(job):
            tracked_events.handled += 1
            tracked_events.failures += 1

        events = [
            JobEvent(JobEventType.COMPLETED, 1, 'A', MockJobResult(20, 'A')),
            JobEvent(JobEventType.FAILURE, 2, 'B', 'ERROR'),
            JobEvent(JobEventType.FAILURE, 3, 'C', 'ERROR'),
            JobEvent(JobEventType.FAILURE, 4, 'D', 'ERROR'),
            JobEvent(JobEventType.COMPLETED, 5, 'E', MockJobResult(22, 'B')),
            JobEvent(JobEventType.FAILURE, 6, 'F', 'ERROR'),
            JobEvent(JobEventType.UNDEFINED, 7, 'G', 'OK'),
        ]

        conn = MockRedisPubSubConnection(events)
        listener = JobEventsListener(conn)

        # COMPLETED and FAILURE events are handled on the first call
        # to run(); on the next call only FAILURE events are handled,
        # tracked_events keeps track of both calls
        listener.subscribe(JobEventType.COMPLETED, handle_successful_job)
        listener.subscribe(JobEventType.FAILURE, handle_failed_job)
        listener.run()
        listener.unsubscribe(JobEventType.COMPLETED)
        listener.run()

        self.assertEqual(tracked_events.handled, 10)
        self.assertEqual(tracked_events.ok, 2)
        self.assertEqual(tracked_events.failures, 8)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#     Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging
import threading
import time

import cherrypy
import rq

from grimoirelab_toolkit.datetime import str_to_datetime

from .arthur import Arthur
from .jobs import JobResult
from .utils import JSONEncoder


logger = logging.getLogger(__name__)


def json_encoder(*args, **kwargs):
    """Custom JSON encoder handler"""

    obj = cherrypy.serving.request._json_inner_handler(*args, **kwargs)

    for chunk in JSONEncoder().iterencode(obj):
        yield chunk.encode('utf-8')


class ArthurServer(Arthur):
    """Arthur REST server"""

    def __init__(self, *args, **kwargs):
        if 'writer' in kwargs:
            writer = kwargs.pop('writer')

        super().__init__(*args, **kwargs)

        if writer:
            self.writer_th = threading.Thread(target=self.write_items,
                                              args=(writer, self.items))
        else:
            self.writer_th = None

        cherrypy.engine.subscribe('start', self.start, 100)

    def start(self):
        """Start the server and the writer"""

        super().start()
        if self.writer_th:
            self.writer_th.start()

    @classmethod
    def write_items(cls, writer, items_generator):
        """Write items to the queue

        :param writer: the writer object
        :param items_generator: items to be written in the queue
        """
        while True:
            items = items_generator()
            writer.write(items)
            time.sleep(1)

    @cherrypy.expose
    @cherrypy.tools.json_in()
    def add(self):
        """Add tasks"""

        payload = cherrypy.request.json

        logger.debug("Reading tasks...")
        for task_data in payload['tasks']:
            try:
                category = task_data['category']
                backend_args = task_data['backend_args']
                archive_args = task_data.get('archive', None)
                sched_args = task_data.get('scheduler', None)
            except KeyError as ex:
                logger.error("Task badly formed")
                raise ex

            from_date = backend_args.get('from_date', None)

            if from_date:
                backend_args['from_date'] = str_to_datetime(from_date)

            super().add_task(task_data['task_id'],
                             task_data['backend'],
                             category,
                             backend_args,
                             archive_args=archive_args,
                             sched_args=sched_args)
        logger.debug("Done. Ready to work!")

        return "Tasks added"

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(handler=json_encoder)
    def remove(self):
        """Remove tasks"""

        payload = cherrypy.request.json
        logger.debug("Reading tasks to remove...")

        task_ids = {}

        for task_data in payload['tasks']:
            task_id = task_data['task_id']
            removed = super().remove_task(task_id)
            task_ids[task_id] = removed

        result = {'tasks': task_ids}

        return result

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out(handler=json_encoder)
    def reschedule(self):
        """Re-schedule failed tasks"""

        payload = cherrypy.request.json

        logger.debug("Reading tasks to re-schedule...")
        task_ids = {}

        for task_data in payload['tasks']:
            task_id = task_data['task_id']
            rescheduled = super().reschedule_task(task_id)
            task_ids[task_id] = rescheduled

        result = {'tasks': task_ids}

        return result

    @cherrypy.expose
    @cherrypy.tools.json_out(handler=json_encoder)
    def tasks(self):
        """List tasks"""

        logger.debug("API 'tasks' method called")

        result = [task.to_dict() for task in self._tasks.tasks]
        result = {'tasks': result}

        logger.debug("Tasks registry read")

        return result

    @cherrypy.expose
    @cherrypy.tools.json_out(handler=json_encoder)
    def task(self, task_id):
        """Get info about a task

        :param task_id: id of the task

        :raises NotFoundError: raised when the requested task is not
-           found in the registry
        :raises TaskRegistryError: raised when the requested task is not
            retrieved from the registry
        """
        logger.debug("API 'task' method called for task %s", task_id)

        task = self._tasks.get(task_id)

        jobs = []

        for job in task.jobs:
            job_rq = rq.job.Job.fetch(job.id, connection=self.conn)
            result = job_rq.result

            if isinstance(result, JobResult):
                job_result = result.to_dict()
            else:
                job_result = None

            job = {
                'task_id': task_id,
                'job_id': job_rq.id,
                'job_number': job.number,
                'job_status': job_rq.get_status(),
                'result': job_result
            }

            jobs.append(job)

        task_dict = task.to_dict()
        task_dict['jobs'] = jobs

        logger.debug("Task %s data generated", task_id)

        return task_dict

    @cherrypy.expose
    @cherrypy.tools.json_out(handler=json_encoder)
    def job(self, job_id):
        """Get info about a job"""

        logger.debug("API 'job' method called for job %s", job_id)

        job_rq = rq.job.Job.fetch(job_id, connection=self.conn)
        result = job_rq.result

        job_result = result.to_dict() if isinstance(result, JobResult) else None

        job_log = job_rq.meta.get('log', None)
        job_number = job_rq.meta.get('job_number', None)

        job = {
            'job_id': job_rq.id,
            'job_number': job_number,
            'job_status': job_rq.get_status(),
            'job_description': job_rq.description,
            'created_at': job_rq.created_at,
            'enqueued_at': job_rq.enqueued_at,
            'started_at': job_rq.started_at,
            'ended_at': job_rq.ended_at,
            'timeout': job_rq.timeout,
            'origin': job_rq.origin,
            'result': job_result,
            'log': job_log
        }

        logger.debug("Job %s data generated", job_id)

        return job



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#

import datetime
import os
import os.path
import re
import shutil
import subprocess
import tempfile
import unittest
import unittest.mock

import dateutil
from redis.exceptions import RedisError

from arthur.arthur import Arthur
from arthur.common import ARCHIVES_DEFAULT_PATH
from arthur.errors import AlreadyExistsError, TaskRegistryError
from arthur.tasks import (ArchivingTaskConfig,
                          SchedulingTaskConfig,
                          TaskStatus)

from base import find_empty_redis_database


INVALID_BACKEND_ARGS_ERROR = "Backend_args is not a dict, task .*"
INVALID_FETCH_FROM_ARCHIVE = "'fetch_from_archive' must be a bool; <class 'int'> given"
INVALID_SCHEDULER_PARAM = "unknown .+ task config parameter"
INVALID_SCHEDULER_DELAY = "'delay' must be an int; <class 'str'> given"
INVALID_SCHEDULER_MAX_RETRIES = "'max_retries' must be an int; <class 'str'> given"
INVALID_ARCHIVED_AFTER_INVALID_DATE_ERROR = "'archived_after' is invalid; X is not a valid date"
MISSING_TASK_ID_ERROR = "Missing task_id for task"
MISSING_BACKEND_ERROR = "Missing backend for task .*"
MISSING_CATEGORY_ERROR = "Missing category for task .*"
MISSING_FETCH_FROM_ARCHIVE_PARAM = ".* missing 1 required positional argument: 'fetch_from_archive'"
UNKNOWN_ARCHIVE_PARAMETER = "unknown .* task config parameter"


class TestArthur(unittest.TestCase):
    """Unit tests for Scheduler class"""

    @classmethod
    def setUpClass(cls):
        cls.tmp_path = tempfile.mkdtemp(prefix='arthur_')
        cls.git_path = os.path.join(cls.tmp_path, 'gittest')

        dir = os.path.dirname(os.path.realpath(__file__))
        subprocess.check_call(['tar', '-xzf',
                               os.path.join(dir, 'data/gittest.tar.gz'),
                               '-C', cls.tmp_path])

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmp_path)

    def setUp(self):
        self.conn = find_empty_redis_database()
        self.conn.flushdb()

    def tearDown(self):
        self.conn.flushdb()

    def test_add_task_no_archive(self):
        """Check when a task has no archive params"""

        task_id = "arthur.task"
        backend = "git"
        category = "commit"
        backend_params = {"a": "a", "b": "b"}

        app = Arthur(self.conn, async_mode=False)

        initial_tasks = len(app._tasks.tasks)
        app.add_task(task_id, backend, category, backend_params)
        after_tasks = len(app._tasks.tasks)

        t = app._tasks.tasks[0]

        self.assertEqual(t.task_id, task_id)
        self.assertEqual(t.backend, backend)
        self.assertDictEqual(t.backend_args, backend_params)
        self.assertEqual(t.archiving_cfg, None)
        self.assertEqual(t.scheduling_cfg, None)

        self.assertEqual(initial_tasks, 0)
        self.assertEqual(after_tasks, 1)

    def test_add_task_empty_archive(self):
        """Check when a task has an empty archive"""

        task_id = "arthur.task"
        backend = "git"
        category = "commit"
        backend_params = {"a": "a", "b": "b"}
        archive_params = None

        app = Arthur(self.conn, async_mode=False)

        initial_tasks = len(app._tasks.tasks)
        app.add_task(task_id, backend, category, backend_params, archive_params)
        after_tasks = len(app._tasks.tasks)

        t = app._tasks.tasks[0]

        self.assertEqual(t.task_id, task_id)
        self.assertEqual(t.backend, backend)
        self.assertDictEqual(t.backend_args, backend_params)
        self.assertEqual(t.archiving_cfg, None)
        self.assertEqual(t.scheduling_cfg, None)

        self.assertEqual(initial_tasks, 0)
        self.assertEqual(after_tasks, 1)

    def test_task_unknown_archive_parameter(self):
        """Check whether an exception is thrown when an unknown parameter is in the archive params"""

        task_id = "arthur.task"
        backend = "git"
        category = None
        backend_params = {"a": "a", "b": "b"}
        archive_params = {"unknown": 1, "fetch_from_archive": True, "archived_after": "2010-10-10"}

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, re.compile(UNKNOWN_ARCHIVE_PARAMETER)):
            app.add_task(task_id, backend, category, backend_params, archive_params)

    def test_task_default_archive_path(self):
        """Check whether a default archive path is added when not defined in the archive params"""

        task_id = "arthur.task"
        backend = "git"
        category = "commit"
        backend_params = {"a": "a", "b": "b"}
        archive_params = {
            'fetch_from_archive': True,
            'archived_after': "2010-10-10"
        }

        app = Arthur(self.conn, async_mode=False)
        app.add_task(task_id, backend, category, backend_params, archive_params)

        t = app._tasks.tasks[0]

        archiving_opts = t.archiving_cfg
        expected_dt = datetime.datetime(2010, 10, 10, 0, 0, 0, tzinfo=dateutil.tz.tzutc())
        self.assertIsInstance(archiving_opts, ArchivingTaskConfig)
        self.assertEqual(archiving_opts.archive_path, os.path.expanduser(ARCHIVES_DEFAULT_PATH))
        self.assertEqual(archiving_opts.fetch_from_archive, True)
        self.assertEqual(archiving_opts.archived_after, expected_dt)

    def test_task_wrong_fetch_from_archive(self):
        """Check whether an exception is thrown when fetch_from_archive parameter is not properly set"""

        task_id = "arthur.task"
        backend = "git"
        category = None
        backend_params = {"a": "a", "b": "b"}
        archive_params = {"fetch_from_archive": 100, "archived_after": "2010-10-10"}

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, INVALID_FETCH_FROM_ARCHIVE):
            app.add_task(task_id, backend, category, backend_params, archive_params)

        archive_params = {"archived_after": "2010-10-10"}
        with self.assertRaisesRegex(TypeError, re.compile(MISSING_FETCH_FROM_ARCHIVE_PARAM)):
            app.add_task(task_id, backend, category, backend_params, archive_params)

    def test_task_ignore_archive_after(self):
        """Check whether the archived_after parameter is not set when fetch_from_archive is false"""

        task_id = "arthur.task"
        backend = "git"
        category = None
        backend_params = {"a": "a", "b": "b"}
        archive_params = {"fetch_from_archive": False, "archived_after": "X"}

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, INVALID_ARCHIVED_AFTER_INVALID_DATE_ERROR):
            app.add_task(task_id, backend, category,
                         backend_params, archive_params)

        self.assertEqual(len(app._tasks.tasks), 0)

    def test_task_wrong_archive_after(self):
        """Check whether an exception is thrown when archived_after parameter is not properly set"""

        task_id = "arthur.task"
        backend = "git"
        category = None
        backend_params = {"a": "a", "b": "b"}
        archive_params = {"fetch_from_archive": True, "archived_after": "X"}

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, INVALID_ARCHIVED_AFTER_INVALID_DATE_ERROR):
            app.add_task(task_id, backend, category, backend_params, archive_params)

    def test_add_task_archive(self):
        """Check whether tasks are added"""

        task_id = "arthur.task"
        backend = "git"
        category = 'acme-product'
        backend_params = {"a": "a", "b": "b"}
        archive_args = {
            'fetch_from_archive': True,
            'archived_after': '2100-01-01'
        }
        sched_args = {
            'delay': 10,
            'max_retries': 5
        }

        app = Arthur(self.conn, base_archive_path=self.tmp_path, async_mode=False)

        initial_tasks = len(app._tasks.tasks)
        app.add_task(task_id, backend, category, backend_params,
                     archive_args, sched_args)
        after_tasks = len(app._tasks.tasks)

        t = app._tasks.tasks[0]

        self.assertEqual(t.task_id, task_id)
        self.assertEqual(t.backend, backend)
        self.assertEqual(t.category, category)
        self.assertDictEqual(t.backend_args, backend_params)

        archiving_opts = t.archiving_cfg
        expected_dt = datetime.datetime(2100, 1, 1, 0, 0, 0, tzinfo=dateutil.tz.tzutc())
        self.assertIsInstance(archiving_opts, ArchivingTaskConfig)
        self.assertEqual(archiving_opts.archive_path, self.tmp_path)
        self.assertEqual(archiving_opts.fetch_from_archive, True)
        self.assertEqual(archiving_opts.archived_after, expected_dt)

        sched_opts = t.scheduling_cfg
        self.assertIsInstance(sched_opts, SchedulingTaskConfig)
        self.assertEqual(sched_opts.delay, 10)
        self.assertEqual(sched_opts.max_retries, 5)

        self.assertEqual(initial_tasks, 0)
        self.assertEqual(after_tasks, 1)

    def test_task_unknown_scheduler_parameter(self):
        """Check whether an exception is thrown when an unknown parameter is in the sched params"""

        task_id = "arthur.task"
        backend = "git"
        category = None
        backend_params = {"a": "a", "b": "b"}
        sched_params = {
            "not_valid_param": 1,
            "delay": 10,
            "max_retries": 10
        }

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, re.compile(INVALID_SCHEDULER_PARAM)):
            app.add_task(task_id, backend, category, backend_params, sched_args=sched_params)

    def test_task_wrong_delay(self):
        """Check whether an exception is thrown when delay parameter is not properly set"""

        task_id = "arthur.task"
        backend = "git"
        category = None
        backend_params = {"a": "a", "b": "b"}
        sched_params = {"delay": "1", "max_retries": 10}

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, INVALID_SCHEDULER_DELAY) as ex:
            app.add_task(task_id, backend, category, backend_params, sched_args=sched_params)

    def test_task_wrong_max_retries(self):
        """Check whether an exception is thrown when max_retries parameter is not properly set"""

        task_id = "arthur.task"
        backend = "git"
        category = None
        backend_params = {"a": "a", "b": "b"}
        sched_params = {"delay": 1, "max_retries": "c"}

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, INVALID_SCHEDULER_MAX_RETRIES) as ex:
            app.add_task(task_id, backend, category, backend_params, sched_args=sched_params)

    def test_add_duplicated_task(self):
        """Check whether an exception is thrown when a duplicated task is added"""

        app = Arthur(self.conn, async_mode=False)

        app.add_task("arthur.task", "git", "commit", {"a": "a", "b": "b"})

        with self.assertRaises(AlreadyExistsError):
            app.add_task("arthur.task", "backend", "category", {"a": "a", "b": "b"})

    @unittest.mock.patch('redis.StrictRedis.set')
    def test_add_task_registry_error(self, mock_redis_set):
        """Check whether a TaskRegistryError exception is thrown when a task cannot be added"""

        mock_redis_set.side_effect = RedisError
        app = Arthur(self.conn, async_mode=False)

        with self.assertRaises(TaskRegistryError):
            app.add_task("arthur.task", "git", "commit", {"a": "a", "b": "b"})

    def test_add_task_no_task_id(self):
        """Check whether an exception is thrown when the task id is missing"""

        task_id = None
        backend = "git"
        category = "commit"
        backend_params = {"a": "a", "b": "b"}

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, MISSING_TASK_ID_ERROR):
            app.add_task(task_id, backend, category, backend_params)

        task_id = "     "
        with self.assertRaisesRegex(ValueError, MISSING_TASK_ID_ERROR):
            app.add_task(task_id, backend, category, backend_params)

    def test_add_task_no_backend(self):
        """Check whether an exception is thrown when the backend is missing"""

        task_id = "arthur.task"
        backend = None
        category = "commit"
        backend_params = {"a": "a", "b": "b"}

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, re.compile(MISSING_BACKEND_ERROR)):
            app.add_task(task_id, backend, category, backend_params)

        backend = "     "
        with self.assertRaisesRegex(ValueError, re.compile(MISSING_BACKEND_ERROR)):
            app.add_task(task_id, backend, category, backend_params)

    def test_add_task_no_category(self):
        """Check whether an exception is thrown when the backend category is missing"""

        task_id = "arthur.task"
        backend = "git"
        category = None
        backend_params = {"a": "a", "b": "b"}

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, re.compile(MISSING_CATEGORY_ERROR)):
            app.add_task(task_id, backend, category, backend_params)

        category = "     "
        with self.assertRaisesRegex(ValueError, re.compile(MISSING_CATEGORY_ERROR)):
            app.add_task(task_id, backend, category, backend_params)

    def test_add_task_invalid_format_backend_args(self):
        """Check whether an exception is thrown when the backend args is not a dict"""

        task_id = "arthur.task"
        backend = "git"
        category = "commit"
        backend_params = "wrong_params"

        app = Arthur(self.conn, async_mode=False)

        with self.assertRaisesRegex(ValueError, re.compile(INVALID_BACKEND_ARGS_ERROR)):
            app.add_task(task_id, backend, category, backend_params)

    def test_remove_task(self):
        """Check whether the removal of tasks is properly handled"""

        task_1 = "arthur.task-1"
        task_2 = "arthur.task-2"

        task_params = {"backend": "git",
                       "category": "commit",
                       "backend_params": {"a": "a", "b": "b"}}

        app = Arthur(self.conn, async_mode=False)

        app.add_task(task_1, task_params['backend'], task_params['category'], task_params['backend_params'])
        app.add_task(task_2, task_params['backend'], task_params['category'], task_params['backend_params'])
        tasks = len(app._tasks.tasks)

        self.assertEqual(tasks, 2)

        self.assertTrue(app.remove_task(task_1))
        self.assertTrue(app.remove_task(task_2))

        tasks = len(app._tasks.tasks)
        self.assertEqual(tasks, 0)

    def test_remove_non_existing_task(self):
        """Check whether the removal of non existing tasks is properly handled"""

        app = Arthur(self.conn, async_mode=False)
        self.assertFalse(app.remove_task("task-x"))

    def test_items(self):
        """Check whether items are properly retrieved"""

        new_path = os.path.join(self.tmp_path, 'newgit')

        app = Arthur(self.conn, async_mode=False)
        app.add_task('test', 'git', 'commit',
                     {'uri': self.git_path,
                      'gitpath': new_path})
        app.start()

        commits = [item['data']['commit'] for item in app.items()]

        expected = ['bc57a9209f096a130dcc5ba7089a8663f758a703',
                    '87783129c3f00d2c81a3a8e585eb86a47e39891a',
                    '7debcf8a2f57f86663809c58b5c07a398be7674c',
                    'c0d66f92a95e31c77be08dc9d0f11a16715d1885',
                    'c6ba8f7a1058db3e6b4bc6f1090e932b107605fb',
                    '589bb080f059834829a2a5955bebfd7c2baa110a',
                    'ce8e0b86a1e9877f42fe9453ede418519115f367',
                    '51a3b654f252210572297f47597b31527c475fb8',
                    '456a68ee1407a77f3e804a30dff245bb6c6b872f']

        self.assertListEqual(commits, expected)

        commits = [item for item in app.items()]

        self.assertListEqual(commits, [])

        shutil.rmtree(new_path)

    def test_reschedule(self):
        """Check whether it reschedules a failed task"""

        task_id = "arthur.task"
        backend = "git"
        category = "commit"
        backend_params = {"a": "a", "b": "b"}
        sched_params = {"max_retries": 5}

        app = Arthur(self.conn, async_mode=False)
        app.add_task(task_id, backend, category,
                     backend_params,
                     sched_args=sched_params)

        # Set task as failed
        t = app._tasks.tasks[0]
        t.status = TaskStatus.FAILED
        t.age = 100
        t.num_failures = 5
        app._tasks.update(task_id, t)

        # Task will be scheduled again after calling the method
        result = app.reschedule_task(task_id)

        # Update task values
        t = app._tasks.get(task_id)

        self.assertEqual(result, True)
        self.assertEqual(t.status, TaskStatus.SCHEDULED)
        self.assertEqual(t.age, 0)
        self.assertEqual(t.num_failures, 0)

    def test_reschedule_non_existing_task(self):
        """Check whether re-scheduling does nothing when the task does not exist"""

        app = Arthur(self.conn, async_mode=False)
        result = app.reschedule_task('no-id')

        self.assertEqual(result, False)
        self.assertEqual(len(app._tasks.tasks), 0)

    def test_reschedule_non_failed_task(self):
        """Check whether the method does nothing when re-scheduling a non failed task"""

        task_id = "arthur.task"
        backend = "git"
        category = "commit"
        backend_params = {"a": "a", "b": "b"}

        app = Arthur(self.conn, async_mode=False)
        app.add_task(task_id, backend, category, backend_params)

        # Force a different status
        t = app._tasks.tasks[0]
        t.status = TaskStatus.RUNNING
        app._tasks.update(task_id, t)

        result = app.reschedule_task('arthur.task')

        # The status should not have changed because
        # the task is a non failed one
        t = app._tasks.get(task_id)
        self.assertEqual(result, False)
        self.assertEqual(t.status, TaskStatus.RUNNING)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Santiago Dueñas <sduenas@bitergia.com>
#

import logging
import unittest

import rq.worker

from arthur.common import CH_PUBSUB
from arthur.events import JobEventType, JobEvent
from arthur.worker import ArthurWorker, JobLogHandler

from base import TestBaseRQ, mock_sum, mock_failure


class MockArthurWorker(rq.worker.SimpleWorker, ArthurWorker):
    """Unit tests for ArthurWorker class, avoids rq inheritance problems"""
    pass


class TestArthurWorker(TestBaseRQ):
    """Unit tests for ArthurWorker class"""

    def test_publish_job_events(self):
        """Tests whether job events are correctly sent"""

        self._publish_job_events(CH_PUBSUB)

    def test_publish_job_events_channel_default(self):
        """Tests whether job events are correctly sent to the default channel"""

        self._publish_job_events(CH_PUBSUB)

    def test_publish_job_events_channel_override(self):
        """"Tests whether job events are correctly sent in a defined channel"""

        self._publish_job_events('test_channel', pubsub_override=True)

    def test_job_logs(self):
        """Tests whether job has logs in their meta field"""

        q = rq.Queue('job_logs_queue')
        w = MockArthurWorker([q])

        job_a = q.enqueue(mock_sum, task_id=0, a=2, b=3)
        job_b = q.enqueue(mock_failure, task_id=1)

        status = w.work(burst=True)
        self.assertEqual(status, True)

        job_a_rq = rq.job.Job.fetch(job_a.id, connection=self.conn)
        job_b_rq = rq.job.Job.fetch(job_b.id, connection=self.conn)

        # Check log list of job 'a'
        self.assertEqual(len(job_a_rq.meta['log']), 5)

        # Check first log message that shows the job was initialized
        self.assertEqual(job_a_rq.meta['log'][0]['module'], 'worker')
        self.assertRegex(job_a_rq.meta['log'][0]['msg'], 'Job OK')

        # Check the last log message that shows when the job finishes
        self.assertEqual(job_a_rq.meta['log'][-1]['module'], 'worker')
        self.assertRegex(job_a_rq.meta['log'][-1]['msg'], 'done, quitting')

        # Check log list of job 'b'
        self.assertEqual(len(job_b_rq.meta['log']), 2)

        # Check first log message that shows the failure of the job
        self.assertEqual(job_b_rq.meta['log'][0]['module'], 'worker')
        self.assertRegex(job_b_rq.meta['log'][0]['msg'], 'Exception')

        # Check the last log message that shows when the job finishes
        self.assertEqual(job_b_rq.meta['log'][-1]['module'], 'worker')
        self.assertRegex(job_b_rq.meta['log'][-1]['msg'], 'done, quitting')

    def _publish_job_events(self, pubsub_channel, pubsub_override=False):
        """Generic job events tests"""

        pubsub = self.conn.pubsub()
        pubsub.subscribe(pubsub_channel)

        q = rq.Queue('foo')
        w = MockArthurWorker([q])

        if pubsub_override:
            w.pubsub_channel = pubsub_channel

        job_a = q.enqueue(mock_sum, task_id=0, a=2, b=3)
        job_b = q.enqueue(mock_failure, task_id=1)

        status = w.work(burst=True)
        self.assertEqual(status, True)

        # Ignore the first messages because it is a
        # subscription notification
        _ = pubsub.get_message()

        # STARTED event for job 'a'
        msg_a = pubsub.get_message()
        event = JobEvent.deserialize(msg_a['data'])
        self.assertEqual(event.job_id, job_a.id)
        self.assertEqual(event.task_id, 0)
        self.assertEqual(event.type, JobEventType.STARTED)
        self.assertEqual(event.payload, None)

        # COMPLETED event for job 'a'
        msg_a = pubsub.get_message()
        event = JobEvent.deserialize(msg_a['data'])
        self.assertEqual(job_a.result, 5)
        self.assertEqual(event.job_id, job_a.id)
        self.assertEqual(event.task_id, 0)
        self.assertEqual(event.type, JobEventType.COMPLETED)
        self.assertEqual(event.payload, 5)

        # STARTED event for job 'b'
        msg_b = pubsub.get_message()
        event = JobEvent.deserialize(msg_b['data'])
        self.assertEqual(event.job_id, job_b.id)
        self.assertEqual(event.task_id, 1)
        self.assertEqual(event.type, JobEventType.STARTED)
        self.assertEqual(event.payload, None)

        # FAILURE event for job 'b'
        msg_b = pubsub.get_message()
        event = JobEvent.deserialize(msg_b['data'])
        self.assertEqual(job_b.result, None)
        self.assertEqual(event.job_id, job_b.id)
        self.assertEqual(event.task_id, 1)
        self.assertEqual(event.type, JobEventType.FAILURE)
        self.assertRegex(event.payload['error'], "Traceback")
        self.assertEqual(event.payload['result'], None)


class TestJobLogHandler(TestBaseRQ):
    """Unit tests for JobLogHandler class"""

    def test_job_log_handler_init(self):
        """Tests whether the handler has initialized well"""

        job_a = rq.job.Job()
        meta_handler = JobLogHandler(job_a)
        self.assertEqual(meta_handler.job, job_a)
        self.assertListEqual(meta_handler.job.meta['log'], [])

    def test_job_log_handler_emit(self):
        """Tests whether the handler catches the messages from the logger that handles"""

        job_a = rq.job.Job()

        # Create handler
        meta_handler = JobLogHandler(job_a)

        # Get logger of this current context and add set level to INFO in order to save info and upper
        logger = logging.getLogger(__name__)
        logger.addHandler(meta_handler)
        logger.setLevel(logging.INFO)

        # Write in the logger
        logger.error("Error log to the handler")
        logger.warning("Warning log to the handler")
        logger.info("Info log to the handler")

        # Check if the logs are saved in the job meta field
        self.assertEqual(len(job_a.meta['log']), 3)
        self.assertEqual(sorted(list(job_a.meta['log'][0].keys())), ['created', 'level', 'module', 'msg'])
        self.assertRegex(job_a.meta['log'][0]['msg'], 'Error')
        self.assertRegex(job_a.meta['log'][-1]['msg'], 'Info')


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arthur', 'arthur.bin', 'tests']

package_data = \
{'': ['*'], 'tests': ['data/*', 'data/redmine/*']}

install_requires = \
['CherryPy>=17.4.2',
 'cheroot>=8.2.1',
 'grimoirelab-toolkit>=0.3',
 'perceval>=0.19',
 'python-dateutil>=2.8.0',
 'redis==3.0.0',
 'rq==1.0.0']

entry_points = \
{'console_scripts': ['arthurd = arthur.bin.arthurd:main',
                     'arthurw = arthur.bin.arthurw:main']}

setup_kwargs = {
    'name': 'kingarthur',
    'version': '0.2.3',
    'description': 'Distributed job queue platform for scheduling Perceval jobs',
    'long_description': '# Arthur [![Build Status](https://github.com/chaoss/grimoirelab-kingarthur/workflows/tests/badge.svg)](https://github.com/chaoss/grimoirelab-kingarthur/actions?query=workflow:tests+branch:master+event:push) [![Coverage Status](https://coveralls.io/repos/github/chaoss/grimoirelab-kingarthur/badge.svg?branch=master)](https://coveralls.io/github/chaoss/grimoirelab-kingarthur?branch=master)\n\nKing Arthur commands his loyal knight Perceval on the quest to fetch\ndata from software repositories.\n\nArthur is a distributed job queue platform that schedules and executes\nPerceval. The platform is composed by two components: `arthurd`, the server\nthat schedules the jobs and one or more instances of `arthurw`, the work horses\nthat will run each Perceval job.\n\nThe repositories whose data will be fetched are added to the\nplatform using a REST API. Then, the server transforms these repositories into\nPerceval jobs and schedules them between its job queues.\n\nWorkers are waiting for new jobs checking these queues. Workers only execute\na job at a time. When a new job arrives, an idle worker will take and run\nit. Once a job is finished, if the result is successful, the server will\nre-schedule it to retrieve new data.\n\nBy default, items fetched by each job will be published using a Redis queue.\nAdditionally, they can be written to an Elastic Search index.\n\n\n## Usage\n\n### arthurd\n```\nusage: arthurd [-c <file>] [-g] [-h <host>] [-p <port>] [-d <database>]\n               [--es-index <index>] [--log-path <path>] [--archive-path <cpath>]\n               [--no-archive] [--no-daemon] | --help\n\nKing Arthur commands his loyal knight Perceval on the quest\nto retrieve data from software repositories.\n\nThis command runs an Arthur daemon that waits for HTTP requests\non port 8080. Repositories to analyze are added using an REST API.\nRepositories are transformed into Perceval jobs that will be\nscheduled and run using a distributed job queue platform.\n\noptional arguments:\n  -?, --help            show this help message and exit\n  -c FILE, --config FILE\n                        set configuration file\n  -g, --debug           set debug mode on\n  -h, --host            set the host name or IP address on which to listen for connections\n  -p, --port            set listening TCP port (default: 8080)\n  -d, --database        URL database connection (default: \'redis://localhost/8\')\n  -s, --sync            work in synchronous mode (without workers)\n  --es-index            output ElasticSearch server index\n  --log-path            path where logs are stored\n  --archive-path        path to archive manager directory\n  --no-archive          do not archive fetched raw data\n  --no-daemon           do not run arthur in daemon mode\n```\n\n#### arthurd configuration file\n\nTo run `arthurd` using a configuration file:\n\n```\n$ arthurd [-c <file>]\n```\n\nWhere `<file>` is the path to an `ini` file which uses the same parameters as in command line, but replacing underscores by hyphens. This configuration file has the following structure:\n\n```\n[arthur]\narchive_path=/tmp/.arthur/archive\ndebug=True\nlog_path=/tmp/logs/arthurd\nno_archive=True\nsync_mode=True\n\n[connection]\nhost=127.0.0.1\nport=8080\n\n[elasticsearch]\nes_index=http://localhost:9200/items\n\n[redis]\ndatabase=redis://localhost/8\n```\n\n### arthurw\n```\nusage: arthurw [-g] [-d <database>] [--burst] [<queue1>...<queueN>] | --help\n\nKing Arthur\'s worker. It will run Perceval jobs on the quest\nto retrieve data from software repositories.\n\npositional arguments:\n   queues               list of queues this worker will listen for\n                        (\'create\' and \'update\', by default)\n\noptional arguments:\n  -?, --help            show this help message and exit\n  -g, --debug           set debug mode on\n  -d, --database        URL database connection (default: \'redis://localhost/8\')\n  -b, --burst           Run in burst mode (quit after all work is done)\n```\n\n## Requirements\n\nKing Arthur requires Python 3.7 or later plus\n[Perceval](https://github.com/chaoss/grimoirelab-perceval/ to run.\nFor other Python dependencies, please check the `pyproject.toml`\nfile included on this repository.\n\nRedis (>= 2.3 and < 3.0) database will also be needed to schedule\nand execute Perceval jobs.\n\n## Installation\n\n```\n$ pip3 install -r requirements.txt\n$ python3 setup.py install\n```\n\n## How to run it\n\nThe first step is to run a Redis server that will be used for communicating\nArthur\'s components. Moreover, an Elastic Search server can be used to store\nthe items generated by jobs. Please refer to their documentation to know how to\ninstall and run them both.\n\nTo run Arthur server:\n```\n$ arthurd -g -d redis://localhost/8 --es-index http://localhost:9200/items --log-path /tmp/logs/arthud --no-archive\n```\n\nTo run a worker:\n\n```\n$ arthurw -d redis://localhost/8\n```\n\n## Adding tasks\n\nTo add tasks to Arthur, create a JSON object containing the tasks needed\nto fetch data from a set of repositories. Each task will run a Perceval\nbackend, thus, backend parameters will also needed for each task.\n\n```\n$ cat tasks.json\n{\n    "tasks": [\n        {\n            "task_id": "arthur.git",\n            "backend": "git",\n            "backend_args": {\n                "gitpath": "/tmp/git/arthur.git/",\n                "uri": "https://github.com/chaoss/grimoirelab-kingarthur.git",\n                "from_date": "2015-03-01"\n            },\n            "category": "commit",\n            "scheduler": {\n                "delay": 10\n            }\n        },\n        {\n            "task_id": "bugzilla_mozilla",\n            "backend": "bugzillarest",\n            "backend_args": {\n                "url": "https://bugzilla.mozilla.org/",\n                "from_date": "2016-09-19"\n            },\n            "category": "bug",\n            "archive": {\n                "fetch_from_archive": true,\n                "archived_after": "2018-02-26 09:00"\n            },\n            "scheduler": {\n                "delay": 60,\n                "max_retries": 5\n            }\n        }\n    ]\n}\n```\n\nThen, send this JSON stream to the server calling `add` method.\n\n```\n$ curl -H "Content-Type: application/json" --data @tasks.json http://127.0.0.1:8080/add\n```\n\nFor this example, items will be stored in the `items` index on the\nElastic Search server (http://localhost:9200/items).\n\n## Listing tasks\n\nThe list of tasks currently scheduled can be obtained using the method `tasks`.\n\n```\n$ curl http://127.0.0.1:8080/tasks\n\n{\n    "tasks": [\n        {\n            "backend_args": {\n                "from_date": "2015-03-01T00:00:00+00:00",\n                "uri": "https://github.com/chaoss/grimoirelab-kingarthur.git",\n                "gitpath": "/tmp/santi/"\n            },\n            "backend": "git",\n            "category": "commit",\n            "created_on": 1480531707.810326,\n            "task_id": "arthur.git",\n            "scheduler": {\n                "max_retries": 3,\n                "delay": 10\n            }\n        }\n    ]\n}\n```\n\n## Removing tasks\n\nScheduled tasks can also be removed calling to the server using the `remove`\nmethod. A JSON stream must be provided setting the identifiers of the\ntasks to be removed.\n\n```\n$ cat tasks_to_remove.json\n\n{\n    "tasks": [\n        {\n            "task_id": "bugzilla_mozilla"\n        },\n        {\n            "task_id": "arthur.git"\n        }\n    ]\n}\n\n$ curl -H "Content-Type: application/json" --data @tasks_to_remove.json http://127.0.0.1:8080/remove\n```\n\n## License\n\nLicensed under GNU General Public License (GPL), version 3 or later.\n',
    'author': 'GrimoireLab Developers',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://chaoss.github.io/grimoirelab/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
